/*****************************************************************************
 *
 *  FileLocator.C                          XMM Science Analysis Software (SAS)
 *
 *  (c) 1997 European Space Agency - written by Marco Beijersbergen
 *
 *****************************************************************************/

#include <fstream>		// C++
#include <strstream>

#include "FileLocator.h"


static void error(const string& mess)
{
  cerr << mess << endl;
  exit(1);
}

static void warning(const string& mess)
{
  cerr << mess << endl;
}



FileLocator::FileLocator() {}

FileLocator::FileLocator(const string& dirList, char separator)
{
  add(dirList,separator);
}


FileLocator::FileLocator(const list<string>& directories)
{
  add(directories);
}

void FileLocator::add(const string& dirList, char separator)
{
  string path = dirList;

  // Split it into directories.
  int prev = 0;
  for (unsigned i = 0; i < dirList.length(); ++i) {
    if (dirList[i] == separator) {
      string dir = dirList.substr(prev,i-prev);
      _directories.push_back(dir);
      prev = i+1;
    }
  }
  string dir = dirList.substr(prev,string::npos);
  _directories.push_back(dir);
}

void FileLocator::add(const list<string>& dirs)
{
  _directories.insert(_directories.end(),dirs.begin(),dirs.end());
}

string FileLocator::locate(const string& fileName, const string& failMessage,
  bool generateError, bool generateWarning) const
{
  string fname(fileName);

  // Handle the empty string case.
  if (fname.empty()) {
    string mess = "Empty filename requested.";
    if (!failMessage.empty()) mess += failMessage;
    if (generateError)
      error(mess);
    else if (generateWarning)
      warning(mess);
    return "";
  }

  // Handle an absolute path.
  if (fname[0] == '/'){
    ifstream ifs(fname.c_str());
    if (ifs) {
      ifs.close();
      // clog << "Using file " << fname << endl;
      return fname;
    }
    else {
      // File not found.
      strstream m;
      m << "Could not open file '" << fname << "' for reading." << endl;
      if (!failMessage.empty())
        m << failMessage <<ends;
      if (generateError)
        error(m.str());
      else if (generateWarning)
         warning(m.str());
      return "";
    }
  }

  // Look for the file in one of these directories.
  for (list<string>::const_iterator i = _directories.begin();
	 i != _directories.end(); ++i) {
    string path = (*i) + "/" + fname;
    ifstream ifs(path.c_str());
    if (ifs) {
      ifs.close();
      // clog << "Using file " << path << endl;
      return path;
    }
  }

  // The file was not found.
  strstream m;
  m << "Could not open file '" << fname << "' for reading in:" << endl;
  for (list<string>::const_iterator i = _directories.begin();
	 i != _directories.end(); ++i)
    m << "  " << (*i) << endl;
  if (!failMessage.empty()) m << failMessage;
  m << ends;
  if (generateError)
     error(m.str());
  else if (generateWarning)
     warning(m.str());
  return "";
}

ostream& operator<<(ostream& os, const FileLocator& f)
{
  for (list<string>::const_iterator it = f._directories.begin();
	 it != f._directories.end(); ++it) {
    if (it != f._directories.begin()) os << ":";
    os << (*it);
  }
  return os;
}

