//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                   (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Random.h:	--- An abstract I/F to a random number generator
//
//	DESCRIPTION:
//		Concrete instances of the provided class Random shall be used
//		to get uniformily distributed pseudo-random numbers in the
//		interval [0, 1].
//
//	CHANGE HISTORY:
//		UL, 1999-05-27: 1.0.0: - first version
//		UL, 1999-06-08: 1.1.0: - added min()/max() methods
//
//	AUTHOR:
//		U. Lammers (UL), Thu May 27 16:15:23 MET DST 1999
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/utils/src/Random.h,v 1.2 1999/06/08 16:08:38 ulammers Exp $
//

#ifndef RANDOM_H
#define RANDOM_H

#include "Tracer.h"

class RandomNumberGenerator {
public:
	enum SophisticationLevel { QUICK_AND_DIRTY, GOOD, REALLY_GOOD,
							   ALMOST_PERFECT, PERFECT };

	RandomNumberGenerator(SophisticationLevel level = ALMOST_PERFECT,
						  unsigned seed = 0, double min = 0., double max = 1.)
					 : _level(level), _seed(seed), _min(min), _max(max)
		   { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Create a random number generator object.
		//
		//	ARGUMENTS:
		//		In:
		//			level	: Describes the level of sophistication
		//					  of the generator in terms of quality of the
		//					  returned random numbers, i.e. long periodicities
		//					  and passing of standard statistical tests;
		//					  Concrete implementations of this class are
		//					  free to ignore the SophisticationLevel.
		//			seed	: Seed value controlling the sequence
		//					  of pseudo-random numbers; identical
		//					  seed values are guaranteed to yield
		//					  the same sequence of pseudo random
		//					  numbers
		//			min/max	: lower/upper bound of numbers returned by
		//					  rand() (see below)
		//

	SophisticationLevel sophisticationLevel() const { PROC RETURN _level; }
	unsigned seed() const { PROC RETURN _seed; }
		//
		//	PURPOSE:
		//		Return the current values of the sophistcation level and
		//		seed which are in effect.
		//

	virtual void sophisticationLevel(SophisticationLevel level) = 0;
	virtual void seed(unsigned seed) = 0;
		//
		//	PURPOSE:
		//		Change the sophisticationLevel/seed of the generator
		//

	double min() const { RETURN _min; }
	double max() const { RETURN _max; }
		//
		//	PURPOSE:
		//		Return the current values of the lower/upper range of
		//		numbers returned by rand()


	virtual void min(double minVal) = 0;
	virtual void max(double maxVal) = 0;
		//
		//	PURPOSE:
		//		Define lower/upper bound of numbers returned by rand() below
		//
		//	ARGUMENTS:
		//		In:	minVal/maxVal	: lower/upper bound of numbers returned
		//							  by rand()
		//


	virtual void reset() = 0;
		//
		//	PURPOSE:
		//		Reset the generator to the state it was in right after
		//		construction
		//

	virtual double rand() = 0;
		//
		//	PURPOSE:
		//		Return a uniformily distributed pseudo-random number in the
		//		range [min, max]
		//

	virtual ~RandomNumberGenerator() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		destroy the object constructed with the above constructors
		//

protected:
	SophisticationLevel	_level;
	unsigned			_seed;
	double				_min;
	double				_max;
};

#define randGen		randGenFunc()
RandomNumberGenerator &randGenFunc();

#endif
