//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	STime.h	--- defines the STime class that represents a point in time 
//
//	DESCRIPTION:
//		Instances of class <STime> represent valid points in time.
//		Constructors and member functions allow to convert between
//		different representations, including Julian/Modified Julian day
//		numbers and calendar day/time. Basic arithmetics is available
//		in the form of operators +/-/+=/-= which take floating point
//		values as a right operands. These are interpreted as time offsets
//		in seconds. The subraction of two STime objects t1 and t2 yields
//		the time difference in seconds between t1 and t2.
//		Note: The class Xmm (declard in Xmm.h of packages caloalutils)
//		defines a mission-wide reference time which is a static STime
//		instance.
//
//		Manipulation of time instances are guarenteed to maintain
//		precision to the micro-sec level.
//
//		There are various constructors to construct an STime object from either
//		a date/time string, a year/month/day/hour/min/secs tuple or
//		a JD/MJD number. If the construction fails due to format errors
//		or out-of-range conditions an InvalidTime() either and exception is
//		thrown or an error message is printed and execution is terminated
//		(depending on setting of STime::errorHandling - see below).
//		In the former case the exception must be caught in the calling
//		context, alternatively, the	success of the operation can be checked
//		a-priori with corresponding	boolean isValidTime() methods (see below).
//		A failed STime construction yields STime::earliest.
//
//	USAGE EXAMPLES:
//		#include <cassert>
//		#include <string>
//		#include <iostream.h>
//		#include "STime.h"
//		#include "Xmm.h"
//
//		STime t1("1997-10-23T14:15:34.123"),	// define time t1
//			  t2 = t1 + 10.,					// t1 10 seconds later
//			  tm = Xmm::missionReferenceTime + 156.1,
//												// 156 seconds after ref time
//			  t3, t4;
//
//		t3 = t2;								/// copy t2 to t3
//		t3 -= 10. * STime::secondsPerDay;		// t3 10 days earlier
//		double diff = t2 - t3;
//
//		cout<<"difference between dates "<<t2<< " and "
//			 <<t3<<" is "<<diff<<" seconds.\n";
//
//		cout<<"Modified Julian day number corresponding to "<<
//			t3.toString()<<" is "<<t3.modifiedJulianDay().full()<<endl;
//
//		cout<<"Mission reference STime is: "<<XMM::missionReferenceTime
//			 << "\n";
//		cout<<"10.3 seconds later is "<<
//				XMM::missionReferenceTime + 10.3<<endl;
//		assert(XMM::missionReferenceTime >= STime::earliest);
//		cout<<"The current time is "<<STime::currentTime()<<endl;
//		cout<<"The current time is "<<STime("now")<<endl;
//
//	CHANGE HISTORY:
//		UL, 1997-10-23: 0.1.0: - first version
//		UL, 1997-10-24: 0.2.0: - comments from JB and MB
//		UL, 1997-12-02: 0.3.0: - bug fix; micro-sec precision
//		UL, 1998-06-13: 1.0.0: - code overhauled
//							   - leap seconds are correctly handled
//		UL, 1998-08-19: 1.1.0: - operator+/operator- return no longer
//								 references
//		UL, 1999-08-05: 1.2.0: - added exception classes to STime
//		UL, 1999-08-09: 1.3.0: - added static isValidTime() methods
//							   - added more explanations
//		UL, 1999-08-09: 1.4.0: - added error handler switch (see below)
//		UL, 1999-08-12: 1.5.0: - fixed typos in description
//							   - removed redundant argument of STime
//								 constructor
//							   - added an optional argument to STime::toString
//								 which determines the number of decimal digits
//								 of the fractional seconds
//		UL, 2000-02-29: 1.6.0: - STime::julian/modifiedJulianDay now returns
//								 JulianDayNumber object
//		UL, 2000-07-04: 1.7.0: - allow construction from Decimal Year Number
//		UL, 2000-11-01: 1.8.0: - added STime::siderialTime()
//
//	AUTHOR:
//		U. Lammers, Thu Oct 23 14:15:34 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/utils/src/STime.h,v 1.16 2001/01/19 11:37:18 ulammers Exp $
///

#ifndef STIME_H
#define STIME_H

#include <iostream>
#include <string>
#include <ctime>
#include <vector>

#include "StringIntMap.h"

//
//	----------------------------( STime )------------------------------------
//
class STime {
public:
	//
	//	exception classes
	//
	//	PURPOSE:
	//		Exception classes to be thrown if construction of STime object
	//		from time string fails (see also isValidTime() below)
	//
	class InvalidTime {
	public:
		InvalidTime(const string &s) : _s(s) {}
		virtual void error() { cerr<<"`"<<_s<<"' is not a valid date/time\n"; }
		const string &arg() const { return _s; }
		virtual ~InvalidTime() {}
	private:
		string _s;
	};
	class InvalidFormat : public InvalidTime {
	public:
		InvalidFormat(const string &s) : InvalidTime(s) {}
	};
	class InvalidRange : public InvalidTime {
	public:
		InvalidRange(const string &s) : InvalidTime(s) {}
		void error() { cerr<<"One or more numeric fields in `"<<arg()
						 <<"' are out of range\n"; }
	};
	class WrongWeekDay : public InvalidTime {
	public:
		WrongWeekDay(const string &s) : InvalidTime(s) {}
		void error() { cerr<<"Invalid week day in `"<<arg()<<"'\n"; }
	};

	enum ErrorHandling { EXCEPTION, ERRORANDEXIT };
	static ErrorHandling errorHandling; // set to ERRORANDEXIT
		//
		//	This determines the way format and out-of-range conditions
		//	are treated in the STime constructors below:
		//		EXCEPTION:	  one of the above exception classes is thrown
		//					  and the constructed object is set to
		//					  STime::earliest; Please note: If this type
		//					  of error handling is chosen possible exceptions
		//					  MUST be caught in the calling context otherwise
		//					  failed STime constructions are likely to
		//					  simply cause the program to core-dump without
		//					  any further messages (implementation dependant)
		//		ERRORANDEXIT: an error message is printed and execution
		//					  is terminated
		//

	STime();
		//
		//	PURPOSE:
		//		Instantiate a STime object set to STime::earliest, i.e.,
		//			STime t; t == STime::earliest
		//		yields true
		//


	STime(unsigned year, unsigned month = 1, unsigned day = 1,
		  unsigned hour = 0, unsigned minute = 0, unsigned secs = 0,
		  double excessSecs = 0.);
		//
		//	PURPOSE:
		//		Instantiate a STime object from a
		//		{year/month/day/hour/minute/secs/excessSecs} tuple
		//		which represents a point in time expressed in UTC.
		//		Passing out-of-range arguments will result in an
		//		InvalidRange() exception being thrown. For checking
		//		if STime construction from
		//		a {year/month/day/hour/minute/secs/excessSecs} tuple will
		//		succeed, use STime::isValidTime() below.
		//
		//  PARAMETERS:
		//      I   : year			: year number [1958, 2029]
		//			  month			: month number [1, 12]
		//			  day			: day number [1, 31]
		//			  hour			: hours since midnight [0, 23]
		//			  minute		: minutes since hour [0, 59]
		//			  secs			: seconds since minute [0, 60] (allow
		//							  one leap second)
		//			  excessSecs	: value in seconds to be added to time -
		//							  normally (not necessarily)
		//							  0 <= excessSecs < 1
		//
		//		Please note: The specified boundaries for the various fields
		//					 are nominal ones but actually given values
		//					 can exceed these bounds. In case they do,
		//					 the number are interpreted as offsets from
		//					 the time instance
		//						 year-01-01T00:00:00
		//					 Example: 1980-14-45T100:111:222 corresponds to
		//					 1981-03-21T05:54:43.000, which is 222 seconds,
		//					 111 minutes, 100 hours, 45 (Julian) days,
		//					 14 months after 1980-01-01T00:00:00.
		//					 


	static bool isValidTime(unsigned year, unsigned month = 1,
							unsigned day = 1, unsigned hour = 0,
							unsigned minute = 0, unsigned secs = 0,
							double excessSecs = 0.);
		//
		//	PURPOSE:
		//		Check if construction of STime object from 
		//		{year/month/day/hour/minute/secs/excessSecs} tuple will
		//		succeed. It will be verified if all arguments are
		//		within the above specified limits and if dates are
		//		valid (example of invalid date is "2000-02-30").
		//


	STime(const string &dateTimeString);
		//
		//	PURPOSE:
		//		Instantiate a STime object from a given time specified
		//		as a string adhereing to either of the formats
		//		FITS/oldFITS/CALENDAR:
		//			- yyyy-mm-dd[Thh:mm:ss[.msec]]  "1997-08-12T23:12:45.1233"
		//			- dd/mm/yy						"31/12/99"
		//			- Wkd Mon day hh:mm::ss yyyy	"Sun Sep 16 01:03:52 1973"
		//		An InvalidFormat() or InvalidRange() exception will be thrown
		//		if the passed time string is invalid or one or more numeric
		//		fields are out of range
		//		(see above for allowd limits). In this the object will
		//		be set to STime::earliest.
		//		The construct STime("now") yields an STime object corresponding
		//		to the current time.
		//
		//  PARAMETERS:
		//		I	: dateTimeString: time string in format according to
		//							  <format> parameter or "now" or "NOW"
		//							  in the latter to cases the STime
		//							  object represents the current time
		//							  (alias for STime::currentTime() - see
		//							  below)
		//							  please note: the evolution around
		//							  the introduction of a leap second is:
		//								1997-06-30T23:59:59.3	- before
		//								1997-06-30T23:59:60.3	- during
		//								1997-07-01T00:00:00.3	- after
		//							  at this point in time TAI is 31s
		//							  ahead of UTC;
		//

	static bool isValidTime(const string &dateTimeString);
		//
		//	PURPOSE:
		//		Check if a given string `dateTimeString' is valid, i.e., a
		//		STime construction will succeed.
		//


	enum Representation { JD, MJD, DYN };
	//	JulianDay/Modified Julian Day/Decimal Year Number
	STime(double timeVal, Representation r = STime::MJD);
		//
		//	PURPOSE:
		//		Instantiate a STime object from double value representing a
		//		Julian or Modified Julian day number in UT or a decimal year
		//		number (e.g. 1998.12345). Passing negative values will result
		//		in an InvalidRange() exception being thrown.
		//
		//  PARAMETERS:
		//		I	: timeVal		: time value corresponding to <t>; has to
		//							  be positive
	    //			  r				: switch determining representation of
		//							  <timeVal>
		//

	static bool isValidTime(double timeVal, Representation r = MJD);
		//
		//	PURPOSE:
		//		Check if a given MJD/JD value is valid, i.e., a
		//		STime construction will succeed.
		//


	STime(unsigned long timeValInt, double timeValFrac = 0.,
		  Representation r = STime::MJD);
		//
		//	PURPOSE:
		//		Instantiate a STime object from an integer/double pair
		//		representing a time as a Julian or Modified Julian day
		//		number broken down into integer and fractional part in UT.
		//
		//  PARAMETERS:
		//		I	: timeValInt	: integer part of time value corresponding
		//							  to
		//							  <t>;
		//			  timeValFrac	: fractional part of time value
		//							  corresponding <t>; has to be positive
	    //			  r				: switch determining representation of
		//							  <timeVal>


	STime(const STime &);
		//
		//	PURPOSE:
		//		copy constructor - initialize object from reference to
		//		another one
		//


	//
	//	----------------------( inquiry methods )---------------------------
	//
	unsigned short year() const;	//	year number [1958, 2029]
	unsigned short month() const;	//	month number [1, 12]
	unsigned short day() const;		//	day number [1, 31]
	unsigned short hour() const;	//  hours elapsed since midnight [0, 23]
	unsigned short minute() const;	//	minutes elapsed since hour [0, 59]
	unsigned short seconds() const;	//	seconds since minute [0, 60] (allow
									//	one leap second)
	unsigned short weekday() const; //  0=Sun, 1=Mon, etc.
	double fracSeconds() const;		//	0 <= fracSeconds <= 1
	unsigned short leapSeconds() const;
									//	number of leap seconds

	//
	//	----------------------( member operators )--------------------------
	//
	STime &operator=(const STime &t);	// assignment operator
	STime &operator+=(double s);		// add/subtract time offset ...
	STime &operator-=(double s);		// ... in seconds


	//
	//	---------------------( other member functions )---------------------
	//
	static STime currentTime(bool asGMT = true); // returns current-time object
												 // as GMT (asGMT=true) or
												 // local time (asGMT=false)
												 //

	enum TimeSystem { UTC, TAI, TT };
		//	UTC	: Coordinated Universal Time
		//	TAI : Internatial Atomic Time
		//	TT  : Terrestial Time

	class JulianDayNumber {
	public:
		JulianDayNumber(long day, double fractionalDay = 0.) :
			_day(day), _fractionalDay(fractionalDay) { normalize(); }
		unsigned long day() const { return _day; }
		void day(unsigned long daynr) { _day = daynr; }
		double fractionalDay() const { return _fractionalDay; }
		void fractionalDay(double frac) { _fractionalDay = frac; }
		double full() const { return static_cast<double>(_day) +
															_fractionalDay; }
		~JulianDayNumber() {}

	private:
		unsigned long	_day;
		double			_fractionalDay;

		void normalize();
	};

	JulianDayNumber julianDay(TimeSystem timeSystem = UTC) const;
										// julian day number
	JulianDayNumber modifiedJulianDay(TimeSystem timeSystem = UTC) const;
										// modified julian day number

	double siderialTime(double longitude = 0.) const;
		//
		//	PURPOSE:
		//		Compute the siderial time at a given longitude on the
		//		surface of the Earth; unit is rad, i.e., 2Pi=24h
		//
		//	ARGUMENTS:
		//		longitude:	longitude of the point on the Earth for which
		//					siderial time is to be computed; units: rad
		//					longitude is 0 on the Greenwich-Meridian,
		//					positively increasing eastwards
		//


	enum StringRepresentation { FITS,	 // "yyyy-mm-dd[Thh:mm:ss[.msec]]"
										 // or "dd/mm/yy", e.g.
										 // 1997-08-12T23:12:45.1233223
								CALENDAR // "Sun Sep 16 01:03:52[.msec] 1973"
					 };
	string toString(unsigned fracSecondDigits = 3,
					StringRepresentation form = STime::FITS) const;
		//	convert to string representation according to formats specified
		//	above; default is FITS format;
		//	<fracSecondDigits> determines the number of digits the fractional
		//	seconds part is printed with; fracSecondDigits = 0 also suppresses
		//	the decimal point.

	~STime();
		//
		//	PURPOSE:
		//		destructor - destroy object constructed with any of the
		//		above constructors


	//
	//	-----------------------------( constants )-------------------------
	//
	static const double secondsPerMinute = 60.;
	static const double secondsPerHour   = 60. * secondsPerMinute;
	static const double secondsPerDay    = 24. * secondsPerHour;

	static const STime	earliest;	// no time earlier than this
	static const STime	latest;		// no time later than this


	//
	//	----------------------------( operators )---------------------------
	//
	friend double operator-(const STime &t1, const STime &t2);
									// t2 - t1 in seconds
	friend bool operator>(const STime &t1, const STime &t2);
									// t1 > t2 if t1 is later
	friend bool operator>=(const STime &t1, const STime &t2);
									// t1 >= t2 if t1 is later or is same
									// point in time
	friend bool operator<(const STime &t1, const STime &t2);
								 	// t1 < t2 if t1 is earlier
	friend bool operator<=(const STime &t1, const STime &t2);
									// t1 <= t2 if t1 is earlier or is same
									// point in time
	friend bool operator!=(const STime &t1, const STime &t2);
									// t1 and t2 are different times
	friend bool operator==(const STime &t1, const STime &t2);
									// t1 and t2 are the same points in time

private:
	//
	//	member functions
	//
	void normalize();				// normalize time
	// initialize data members
	void	initMembers(unsigned, unsigned, unsigned, unsigned, unsigned,
						unsigned, double);
	void	initMembers(unsigned long, double);
	double	toJulianDay(double &JDfrac);
	static bool	isWithinLimits(unsigned year, unsigned month, unsigned day,
							   unsigned hour, unsigned minute, unsigned secs,
							   double excessSecs);
	static unsigned leapYear(unsigned year);
	string formatTime(unsigned fracSecondDigits) const;
	double hoursAsNormalizedArc() const;

	//
	//	data members
	//
	struct tm	_t;				// base time ...
	double		_excessSecs;	// ... plus fractional seconds precision
								// 0 < _excessSecs < 1
	unsigned long
				_modJuldayInt;	// integer part of modified Julian day
	double		_modJuldayFrac;	// fractional part of modified Julian day;
	unsigned short
				_nLeapSeconds;	// leap seconds

	static StringIntMap<int> _monthNames,
							 _weekDayNames;
};


//
//	------------( non-member operators/functions )--------------------------
//
STime operator+(const STime &t, double s);	// add a number of seconds to
											// a STime
STime operator-(const STime &t, double s);	// subtrcat number of seconds
											// from a time
double operator-(const STime &t1, const STime &t2);
											// difference in secs between
											// two times
ostream &operator<<(ostream &os, const STime &t);
											// print a time using "<<"
											// notation

//
//	-----------------( relational operators - see above )-------------------
bool operator>(const STime &t1, const STime &t2);
bool operator>=(const STime &t1, const STime &t2);
bool operator<(const STime &t1, const STime &t2);
bool operator<=(const STime &t1, const STime &t2);
bool operator!=(const STime &t1, const STime &t2);
bool operator==(const STime &t1, const STime &t2);


//
//	-----------------------------( LeapSeconds )-----------------------------
//
typedef pair<double, unsigned short> LeapPair;	//	[MJD, TAI-UTC]
class LeapSeconds : public vector<LeapPair> {
public:
	LeapSeconds(const string &leapSecondsFile = "");
		//
		//	PURPOSE:
		//		instantiate LeapSeconds object from contents of
		//		leap seconds file
		//
		//	DESCRIPTION:
		//		The object is initialized with the contents of the
		//		specified data set containing leap second information;
		//		this can either stem from the FTOOLS refdata sections
		//		(ftp://legacy.gsfc.nasa.gov:/caldb/data/gen/bcf/leapsec*.fits)
		//		or be an instance of the the XMM CCF::LeapSeconds
		//		constituent. If no data set name is specified, the
		//		environment variable
		//			LEAP_FILE
		//		is supposed to point to the corresponding data set. If this
		//		environment variable is not set or does not point to
		//		a valid data set, the object is initialized from the list of
		//		known leap seconds till 1/7/97.
		//


	unsigned short leapSeconds(double modifiedJulianDayNumber) const;
		//
		//	PRUPOSE:
		//		compute the number of leap seconds introduced since
		//		a point in time expressed as a modified Julian day number
		//		in UT.
		//
		//	PARAMETERS:
		//		I:	modifiedJulianDayNumber: in UT
		//
		//	RETURNS:
		//		number of leap seconds introduced since
		//		<modifiedJulianDayNumber>
		//


	short howManyBetween(double modifiedJulianDayNumber,
						 double deltaTinSecs) const;
		//
		//	PURPOSE:
		//		compute the number of leap seconds that where introduced
		//		over a time span <deltaTinSecs> counting from
		//		a point in time expressed in modified Julian days (UT)
		//
		//	PARAMETERS:
		//		I:	modifiedJulianDayNumber	: in UT
		//			deltaTinSecs			: time offset in SI seconds
		//
		//	RETURNS:
		//		number of leap seconds introduces over time period
		//		deltaTinSecs
		//


	~LeapSeconds();
		//
		//	PURPOSE:
		//		destructs objects created with any of the above constructors
		//
};

#endif
