//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	SimpleToken.cc	--- implements the SimpleToken class which represents
//                          basic string tokens.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/utils/src/SimpleToken.cc,v 1.3 2000/02/22 14:35:37 mthomas Exp $
///


#include <map>
#include <string>

#include "Tracer.h"
#include "SimpleToken.h"


class SimpleToken::Value
{
	public :

		virtual ~Value() {}
		virtual char asChar() const = 0;
		virtual string asString() const = 0;

	protected :
		void invalidOperation() const { PROC cerr << "invalid operation" << endl; exit( 1 ); }
};


class CharValue : public virtual SimpleToken::Value
{
	public :
		CharValue( char v ) : _value( v) { PROC ENDPROC }
		char asChar() const { PROC RETURN _value; }
		string asString() const { PROC invalidOperation(); RETURN ""; }
	private :
		char	_value;
};



class StringValue : public virtual SimpleToken::Value
{
	public :
		StringValue( const string& v ) : _value( v ) { PROC ENDPROC }
		string asString() const { PROC RETURN _value; }
		char asChar() const { PROC invalidOperation(); RETURN ' '; }
	private :
		string	_value;
};


SimpleToken::SimpleToken() : _value(0), _id( SimpleToken::Undefined )
{
	PROC
	ENDPROC
}


SimpleToken::SimpleToken( const SimpleToken& org ) : _value( 0 ), _id( org.id() )
{
	PROC
	switch( _id )
	{
		case SimpleToken::Undefined :
			break;
		case SimpleToken::Character :
			_value = new CharValue( org.asChar() );
			break;
		case SimpleToken::String :
			_value = new StringValue( org.asString() );
			break;
		case SimpleToken::Eof :
			break;
		default :
			cerr << "Unhandled case" << endl;
			exit( 1 );
	} // switch
	ENDPROC
}


SimpleToken&
SimpleToken::operator=( const SimpleToken& org )
{
	PROC
	delete _value;
	_value = 0;
	_id = org.id();
	switch( _id )
	{
		case SimpleToken::Undefined :
			break;
		case SimpleToken::Character :
			_value = new CharValue( org.asChar() );
			break;
		case SimpleToken::String :
			_value = new StringValue( org.asString() );
			break;
		case SimpleToken::Eof :
			break;
		default :
			cerr << "Unhandled case" << endl;
			exit( 1 );
	} // switch

	RETURN *this;
}


SimpleToken::SimpleToken( char v ) : _value( new CharValue( v ) ), _id( SimpleToken::Character )
{
	PROC
	ENDPROC
}


SimpleToken::SimpleToken( const string& v ) : _value( new StringValue( v ) ), _id( SimpleToken::String )
{
	PROC
	ENDPROC
}


SimpleToken::SimpleToken( Id id ) : _value( 0 ), _id( id )
{
	PROC
	switch( _id )
	{
		case SimpleToken::Eof :
		case SimpleToken::Undefined :
			break;
		case SimpleToken::Character :
			_value = new CharValue( '\0' );
			break;
		case SimpleToken::String :
			_value = new StringValue( "" );
			break;
		default :
			cerr << "Unhandled case" << endl; exit( 1 );
			break;
	} // switch
	ENDPROC
}


SimpleToken::~SimpleToken()
{
	PROC
	delete _value;
	_value = 0;
	ENDPROC
}


SimpleToken::Id
SimpleToken::id() const
{
	PROC
	RETURN _id;
}


void
SimpleToken::id( Id id )
{
	PROC
	*this = SimpleToken( id );
	ENDPROC
}


void
SimpleToken::value( char v )
{
	PROC
	*this = SimpleToken( v );
	ENDPROC
}


void
SimpleToken::value( const string& v )
{
	PROC
	*this = SimpleToken( v );
	ENDPROC
}


char
SimpleToken::asChar() const
{
	PROC
	RETURN _value -> asChar();
}


string
SimpleToken::asString() const
{
	PROC
	RETURN _value -> asString();
}


SimpleToken&
SimpleToken::operator=( char v )
{
	PROC
	delete _value;
	_id = SimpleToken::Character;
	_value = new CharValue( v );
	RETURN *this;
}


SimpleToken&
SimpleToken::operator=( const string& v )
{
	PROC
	delete _value;
	_id = SimpleToken::String;
	_value = new StringValue( v );
	RETURN *this;
}


SimpleToken&
SimpleToken::operator=( Id id )
{
	PROC

	delete _value;
	_value = 0;
	_id = id;

	switch( _id )
	{
		case SimpleToken::Eof :
		case SimpleToken::Undefined :
			break;
		case SimpleToken::Character :
			_value = new CharValue( '\0' );
			break;
		case SimpleToken::String :
			_value = new StringValue( "" );
			break;
		default :
			cerr << "Unhandled case" << endl; exit( 1 );
	} // switch

	RETURN *this;
}


static struct
{
	SimpleToken::Id	id;
	char*		str;
} ids[] = {
	{ SimpleToken::Undefined, 	"Undefined" },
	{ SimpleToken::Character, 	"Character" },
	{ SimpleToken::String, 		"String" },
	{ SimpleToken::Eof, 		"Eof" }};


string
SimpleToken::idAsString() const
{

	static bool				firstTime = true;
	static map< SimpleToken::Id, char*>	stringOf;

	if( firstTime )
	{
		const unsigned int nr = sizeof( ids ) / sizeof( ids[ 0 ] );

		for( unsigned int i = 0; i < nr; ++i )
		{
			stringOf[ ids[ i ].id ] = ids[ i ].str;
		} // for
		firstTime = false;
	} // if
	RETURN stringOf[ _id ];
}


